#include <algorithm>
#include <limits>
#include <msp/core/maputils.h>
#include <msp/strings/format.h>
#include "arrangement.h"
#include "container.h"
#include "layout.h"
#include "widget.h"

using namespace std;

namespace Msp {
namespace GLtk {

class Layout::LinearProgram
{
public:
	class Row
	{
	private:
		LinearProgram &linprog;
		unsigned index;

	public:
		Row(LinearProgram &, unsigned);

		float &operator[](unsigned);
		float &back();
	};

private:
	struct Column
	{
		unsigned basic;
		std::vector<float> values;

		Column();
	};

	unsigned n_columns;
	unsigned n_rows;
	std::vector<Column> columns;
	bool solved;
	bool infeasible;

public:
	LinearProgram(unsigned);

	Row add_row();
	Row operator[](unsigned);
	Row get_objective_row();
	float get_variable(unsigned);
	bool solve();
private:
	void prepare_columns();
	void add_artificial_variables();
	void remove_artificial_variables();
	unsigned find_minimal_ratio(unsigned);
	void make_basic_column(unsigned, unsigned);
	bool pivot();
};


struct Layout::Pointers
{
	int Geometry::*pos;
	unsigned Geometry::*dim;
	Packing Slot::*packing;
	unsigned Sides::*low_margin;
	unsigned Sides::*high_margin;
	unsigned Layout::*spacing;
};

Layout::Pointers Layout::pointers[2] =
{ {
		&Geometry::x, &Geometry::w,
		&Slot::horiz_pack,
		&Sides::left, &Sides::right,
		&Layout::col_spacing
	}, {
		&Geometry::y, &Geometry::h,
		&Slot::vert_pack,
		&Sides::bottom, &Sides::top,
		&Layout::row_spacing
} };


Layout::Layout():
	container(0),
	n_active_slots(0),
	margin(8),
	row_spacing(5),
	col_spacing(4)
{
	n_slack_constraints[0] = 0;
	n_slack_constraints[1] = 0;
}

Layout::~Layout()
{
	for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
		delete *i;
}

void Layout::set_container(Container &c)
{
	if(container)
		throw logic_error("container!=0");

	container = &c;
}

void Layout::set_margin(const Sides &m)
{
	margin = m;
	if(container)
		update();
}

void Layout::set_spacing(unsigned s)
{
	row_spacing = s;
	col_spacing = s;
	if(container)
		update();
}

void Layout::set_row_spacing(unsigned s)
{
	row_spacing = s;
	if(container)
		update();
}

void Layout::set_column_spacing(unsigned s)
{
	col_spacing = s;
	if(container)
		update();
}

void Layout::push_arrangement(Arrangement &arr)
{
	arrangement_stack.push_back(&arr);
}

Arrangement *Layout::get_arrangement() const
{
	if(arrangement_stack.empty())
		return 0;
	else
		return arrangement_stack.back();
}

void Layout::pop_arrangement(Arrangement &arr)
{
	list<Arrangement *>::iterator begin = find(arrangement_stack.begin(), arrangement_stack.end(), &arr);
	if(begin==arrangement_stack.end())
		return;

	while(1)
	{
		Arrangement *top = arrangement_stack.back();
		arrangement_stack.pop_back();
		if(!arrangement_stack.empty())
			arrangement_stack.back()->arrange(*top);
		if(top==&arr)
			break;
	}
}

void Layout::add_widget(Widget &wdg)
{
	if(!container)
		throw logic_error("!container");

	slots.push_back(new Slot(*this, wdg));
	update_slot_indices();
	if(!arrangement_stack.empty())
		arrangement_stack.back()->arrange(wdg);
	if(container)
		update();
}

void Layout::remove_widget(Widget &wdg)
{
	for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
		if(&(*i)->widget==&wdg)
		{
			for(list<Slot *>::iterator j=slots.begin(); j!=slots.end(); ++j)
				if(j!=i)
				{
					for(list<Constraint>::iterator k=(*j)->constraints.begin(); k!=(*j)->constraints.end(); )
					{
						if(&k->target==*i)
							(*j)->constraints.erase(k++);
						else
							++k;
					}
				}

			delete *i;
			slots.erase(i);

			update_slot_indices();
			update();
			return;
		}
}

void Layout::update_slot_indices()
{
	n_active_slots = 0;
	for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
	{
		if((*i)->widget.is_visible() || (*i)->ghost)
			(*i)->index = n_active_slots++;
		else
			(*i)->index = -1;
	}

	n_slack_constraints[0] = 0;
	n_slack_constraints[1] = 0;
	for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
		if((*i)->index>=0)
		{
			for(list<Constraint>::iterator j=(*i)->constraints.begin(); j!=(*i)->constraints.end(); ++j)
				if(j->target.index>(*i)->index && (j->type&SLACK))
					++n_slack_constraints[j->type&1];
		}
}

Layout::Slot &Layout::get_slot_for_widget(Widget &wdg)
{
	for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
		if(&(*i)->widget==&wdg)
			return **i;

	throw hierarchy_error("widget not in layout");
}

Layout::ConstraintType Layout::complement(ConstraintType type)
{
	return static_cast<ConstraintType>((type&~(SELF_MASK|TARGET_MASK)) | ((type&SELF_MASK)<<2) | ((type&TARGET_MASK)>>2));
}

void Layout::create_constraint(Widget &src, ConstraintType type, Widget &tgt, int sp)
{
	if(&src==&tgt)
		throw invalid_argument("&src==&tgt");

	Slot &src_slot = get_slot_for_widget(src);
	Slot &tgt_slot = get_slot_for_widget(tgt);

	for(list<Constraint>::iterator i=src_slot.constraints.begin(); i!=src_slot.constraints.end(); ++i)
		if(i->type==type && &i->target==&tgt_slot)
			return;

	src_slot.constraints.push_back(Constraint(type, tgt_slot));
	src_slot.constraints.back().spacing = sp;
	tgt_slot.constraints.push_back(Constraint(complement(type), src_slot));
	tgt_slot.constraints.back().spacing = sp;

	update_slot_indices();
	update();
}

void Layout::add_constraint(Widget &src, ConstraintType type, Widget &tgt)
{
	create_constraint(src, type, tgt, -1);
}

void Layout::add_constraint(Widget &src, ConstraintType type, Widget &tgt, unsigned spacing)
{
	create_constraint(src, type, tgt, spacing);
}

void Layout::set_gravity(Widget &wdg, int h, int v)
{
	Slot &slot = get_slot_for_widget(wdg);

	slot.horiz_pack.gravity = h;
	slot.vert_pack.gravity = v;

	update();
}

void Layout::set_expand(Widget &wdg, bool h, bool v)
{
	Slot &slot = get_slot_for_widget(wdg);

	slot.horiz_pack.expand = h;
	slot.vert_pack.expand = v;

	update();
}

void Layout::set_ghost(Widget &wdg, bool g)
{
	Slot &slot = get_slot_for_widget(wdg);

	slot.ghost = g;

	if(!wdg.is_visible())
	{
		update_slot_indices();
		update();
	}
}

void Layout::update()
{
	solve_constraints(HORIZONTAL, UPDATE);
	solve_constraints(VERTICAL, UPDATE);

	for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
		(*i)->widget.set_geometry((*i)->geom);
}

void Layout::autosize()
{
	solve_constraints(HORIZONTAL, AUTOSIZE);
	solve_constraints(VERTICAL, AUTOSIZE);

	container->set_size(autosize_geom.w, autosize_geom.h);
}

void Layout::solve_constraints(int dir, SolveMode mode)
{
	Pointers &ptrs = pointers[dir&VERTICAL];

	const Geometry &geom = container->get_geometry();
	if(mode==UPDATE && geom.*(ptrs.dim)<margin.*(ptrs.low_margin)+margin.*(ptrs.high_margin))
		return;

	/* Set up a linear program to solve the constraints.  The program matrix has
	five columns for each widget, and one constant column.  The first and second
	columns of a widget are its position and dimension, respectively.  The
	remaining three are slack columns; see below for their purposes. */
	LinearProgram linprog(n_active_slots*5+n_slack_constraints[dir]+1);
	float weight = slots.size();
	for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
	{
		if((*i)->index<0)
			continue;

		LinearProgram::Row objective = linprog.get_objective_row();
		if(mode==AUTOSIZE)
		{
			objective[(*i)->index*5] = -1;
			objective[(*i)->index*5+1] = -1;
		}
		else
		{
			objective[(*i)->index*5] = ((*i)->*(ptrs.packing)).gravity/weight;
			objective[(*i)->index*5+1] = (((*i)->*(ptrs.packing)).expand ? weight : -1);
		}

		{
			// Prevent the widget from going past the container's low edge.
			LinearProgram::Row row = linprog.add_row();
			row[(*i)->index*5] = 1;
			row[(*i)->index*5+2] = -1;
			row.back() = margin.*(ptrs.low_margin);
		}

		if(mode==UPDATE)
		{
			// Prevent the widget from going past the container's high edge.
			LinearProgram::Row row = linprog.add_row();
			row[(*i)->index*5] = 1;
			row[(*i)->index*5+1] = 1;
			row[(*i)->index*5+3] = 1;
			row.back() = geom.*(ptrs.dim)-margin.*(ptrs.high_margin);
		}

		if(((*i)->*(ptrs.packing)).gravity==0)
		{
			/* This forces the widget's distance from the left and right edge of
			the container to be equal.  It's a bit of a hack, but more time and
			thought is needed for a better solution. */
			LinearProgram::Row row = linprog.add_row();
			row[(*i)->index*5+2] = 1;
			row[(*i)->index*5+3] = -1;
		}

		{
			/* Don't allow the widget's dimension to get below that determined
			by autosizing. */
			LinearProgram::Row row = linprog.add_row();
			row[(*i)->index*5+1] = 1;
			row[(*i)->index*5+4] = -1;
			row.back() = (*i)->autosize_geom.*(ptrs.dim);
		}

		/* Add rows for user-defined constraints.  Constraints are always added
		in pairs, so it's only necessary to create a row for one half. */
		unsigned k = n_active_slots*5;
		for(list<Constraint>::iterator j=(*i)->constraints.begin(); j!=(*i)->constraints.end(); ++j)
			if(j->target.index>(*i)->index && (j->type&1)==dir)
			{
				LinearProgram::Row row = linprog.add_row();
				float polarity = ((j->type&SELF_DIM) ? -1 : 1);
				if(j->type&SELF_POS)
					row[(*i)->index*5] = polarity;
				if(j->type&SELF_DIM)
					row[(*i)->index*5+1] = polarity;
				if(j->type&TARGET_POS)
					row[j->target.index*5] = -polarity;
				if(j->type&TARGET_DIM)
					row[j->target.index*5+1] = -polarity;
				if(j->type&SPACING)
					row.back() = (j->spacing>=0 ? j->spacing : this->*(ptrs.spacing));
				if(j->type&SLACK)
					row[k++] = -1;
			}
	}

	if(!linprog.solve())
		return;

	if(mode==AUTOSIZE)
	{
		autosize_geom.*(ptrs.dim) = 0;
		for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
			if((*i)->index>=0)
			{
				int high_edge = linprog.get_variable((*i)->index*5)+linprog.get_variable((*i)->index*5+1);
				autosize_geom.*(ptrs.dim) = max(autosize_geom.*(ptrs.dim), high_edge+margin.*(ptrs.high_margin));
			}
	}
	else
	{
		for(list<Slot *>::iterator i=slots.begin(); i!=slots.end(); ++i)
			if((*i)->index>=0)
			{
				(*i)->geom.*(ptrs.pos) = linprog.get_variable((*i)->index*5);
				(*i)->geom.*(ptrs.dim) = linprog.get_variable((*i)->index*5+1);
			}
	}
}


Layout::Constraint::Constraint(ConstraintType t, Slot &s):
	type(t),
	target(s),
	spacing(-1)
{ }


Layout::Packing::Packing():
	gravity(-1),
	expand(false)
{ }


Layout::Slot::Slot(Layout &l, Widget &w):
	layout(l),
	index(0),
	widget(w),
	ghost(false)
{
	vert_pack.gravity = 1;
	widget.signal_autosize_changed.connect(sigc::mem_fun(this, &Slot::autosize_changed));
	widget.signal_visibility_changed.connect(sigc::mem_fun(this, &Slot::visibility_changed));
	widget.autosize();
	autosize_geom = widget.get_geometry();
}

void Layout::Slot::autosize_changed()
{
	widget.autosize();
	autosize_geom = widget.get_geometry();

	if(!widget.is_visible() && !ghost)
		return;

	// If the widget fits in the area it had, just leave it there.
	if(autosize_geom.w<=geom.w && autosize_geom.h<=geom.h)
		widget.set_geometry(geom);
	else
	{
		layout.container->signal_autosize_changed.emit();
		layout.update();
	}
}

void Layout::Slot::visibility_changed(bool v)
{
	layout.update_slot_indices();
	if(v || ghost)
	{
		layout.container->signal_autosize_changed.emit();
		layout.update();
	}
}


Layout::Loader::Loader(Layout &l, const WidgetMap &wm):
	DataFile::ObjectLoader<Layout>(l),
	wdg_map(wm)
{
	add("column_spacing", &Loader::column_spacing);
	add("margin",         &Loader::margin);
	add("row_spacing",    &Loader::row_spacing);
	add("spacing",        &Loader::spacing);
	add("widget",         &Loader::widget);
}

void Layout::Loader::column_spacing(unsigned s)
{
	obj.set_column_spacing(s);
}

void Layout::Loader::margin()
{
	Sides sides;
	load_sub(sides);
	obj.set_margin(sides);
}

void Layout::Loader::spacing(unsigned s)
{
	obj.set_spacing(s);
}

void Layout::Loader::row_spacing(unsigned s)
{
	obj.set_row_spacing(s);
}

void Layout::Loader::widget(const string &n)
{
	Widget &wdg = *get_item(wdg_map, n);
	WidgetLoader ldr(obj, wdg, wdg_map);
	load_sub_with(ldr);
}


Layout::WidgetLoader::WidgetLoader(Layout &l, Widget &w, const Layout::Loader::WidgetMap &wm):
	layout(l),
	widget(w),
	wdg_map(wm)
{
	add("constraint", &WidgetLoader::constraint);
	add("expand",     &WidgetLoader::expand);
	add("ghost",      &WidgetLoader::ghost);
	add("gravity",    &WidgetLoader::gravity);
}

void Layout::WidgetLoader::constraint(ConstraintType type, const string &n)
{
	Widget &target = *get_item(wdg_map, n);
	layout.add_constraint(widget, type, target);
}

void Layout::WidgetLoader::expand(bool h, bool v)
{
	layout.set_expand(widget, h, v);
}

void Layout::WidgetLoader::ghost(bool g)
{
	layout.set_ghost(widget, g);
}

void Layout::WidgetLoader::gravity(int h, int v)
{
	layout.set_gravity(widget, h, v);
}


void operator>>(const LexicalConverter &conv, Layout::ConstraintType &ctype)
{
	const string &str = conv.get();
	if(str=="ABOVE")
		ctype = Layout::ABOVE;
	else if(str=="BELOW")
		ctype = Layout::BELOW;
	else if(str=="RIGHT_OF")
		ctype = Layout::RIGHT_OF;
	else if(str=="LEFT_OF")
		ctype = Layout::LEFT_OF;
	else if(str=="FAR_ABOVE")
		ctype = Layout::FAR_ABOVE;
	else if(str=="FAR_BELOW")
		ctype = Layout::FAR_BELOW;
	else if(str=="FAR_RIGHT_OF")
		ctype = Layout::FAR_RIGHT_OF;
	else if(str=="FAR_LEFT_OF")
		ctype = Layout::FAR_LEFT_OF;
	else if(str=="ALIGN_TOP")
		ctype = Layout::ALIGN_TOP;
	else if(str=="ALIGN_BOTTOM")
		ctype = Layout::ALIGN_BOTTOM;
	else if(str=="ALIGN_RIGHT")
		ctype = Layout::ALIGN_RIGHT;
	else if(str=="ALIGN_LEFT")
		ctype = Layout::ALIGN_LEFT;
	else if(str=="COPY_WIDTH")
		ctype = Layout::COPY_WIDTH;
	else if(str=="COPY_HEIGHT")
		ctype = Layout::COPY_HEIGHT;
	else
		throw lexical_error(format("conversion of '%s' to ConstraintType", str));
}


Layout::LinearProgram::LinearProgram(unsigned s):
	n_columns(s),
	n_rows(1),
	columns(n_columns),
	solved(false),
	infeasible(false)
{ }

Layout::LinearProgram::Row Layout::LinearProgram::add_row()
{
	return Row(*this, n_rows++);
}

Layout::LinearProgram::Row Layout::LinearProgram::operator[](unsigned r)
{
	if(r>=n_rows)
		throw out_of_range("LinearProgram::operator[]");

	return Row(*this, r);
}

Layout::LinearProgram::Row Layout::LinearProgram::get_objective_row()
{
	return Row(*this, 0);
}

float Layout::LinearProgram::get_variable(unsigned i)
{
	if(!solved || infeasible)
		throw logic_error("not solved");
	if(i+1>=n_columns)
		throw out_of_range("LinearProgram::get_variable");

	if(unsigned r = columns[i].basic)
		return columns.back().values[r];
	else
		return 0;
}

bool Layout::LinearProgram::solve()
{
	if(solved || infeasible)
		return !infeasible;

	/* Solve the program using the simplex method.  The column representing the
	objective variable is kept implicit, as it would never change during the
	execution of the algorithm. */

	prepare_columns();

	add_artificial_variables();

	// Solve the phase 1 problem.
	while(pivot()) ;

	/* All artificial variables should now be non-basic and thus zero, so the
	objective function's value should also be zero.  If it isn't, the original
	program can't be solved. */
	if(columns.back().values.front())
	{
		infeasible = true;
		return false;
	}

	remove_artificial_variables();

	// Solve the phase 2 problem.  We already know it to be feasible.
	while(pivot()) ;

	solved = true;

	return true;
}

void Layout::LinearProgram::prepare_columns()
{
	/* See if any variables are already basic.  A basic variable must only have
	a nonzero coefficient on one row, and its product with the constant column
	must not be negative.  Only one variable can be basic for any given row. */
	vector<float> obj_coeff(n_rows, 0.0f);
	vector<float> row_coeff(n_rows, 1.0f);
	const vector<float> &constants = columns.back().values;
	for(vector<Column>::iterator i=columns.begin(); i!=columns.end(); ++i)
	{
		if(i->values.size()>=2 && i->values.back()!=0.0f && (constants.size()<i->values.size() || i->values.back()*constants[i->values.size()-1]>=0.0f) && obj_coeff[i->values.size()-1]==0.0f)
		{
			bool basic = true;
			for(unsigned j=1; (basic && j+1<i->values.size()); ++j)
				basic = (i->values[j]==0.0f);
			if(basic)
			{
				i->basic = i->values.size()-1;
				row_coeff[i->basic] = 1.0f/i->values.back();
				obj_coeff[i->basic] = -i->values.front();
				i->values.clear();
			}
		}
	}

	// Price out the newly-created basic variables.
	for(vector<Column>::iterator i=columns.begin(); i!=columns.end(); ++i)
		if(!i->values.empty())
		{
			for(unsigned j=0; j<i->values.size(); ++j)
			{
				i->values[j] *= row_coeff[j];
				i->values.front() += obj_coeff[j]*i->values[j];
			}
		}
}

void Layout::LinearProgram::add_artificial_variables()
{
	vector<unsigned> artificial_rows(n_rows-1);
	for(unsigned i=0; i<artificial_rows.size(); ++i)
		artificial_rows[i] = i+1;

	for(vector<Column>::iterator i=columns.begin(); i!=columns.end(); ++i)
		if(i->basic)
			artificial_rows[i->basic-1] = 0;
	artificial_rows.erase(std::remove(artificial_rows.begin(), artificial_rows.end(), 0), artificial_rows.end());

	/* Force all non-basic columns fully into existence and relocate objective
	row to bottom in preparation of phase 1.  A new objective row is calculated
	by pricing out the constraint rows. */
	for(vector<Column>::iterator i=columns.begin(); i!=columns.end(); ++i)
		if(!i->basic)
		{
			float objective = 0.0f;
			if(!i->values.empty())
			{
				objective = i->values.front();
				i->values.front() = 0.0f;
				for(vector<unsigned>::iterator j=artificial_rows.begin(); j!=artificial_rows.end(); ++j)
					if(*j<i->values.size())
						i->values.front() += i->values[*j];
			}
			i->values.resize(n_rows+1, 0.0f);
			i->values.back() = objective;
		}

	if(artificial_rows.empty())
		return;

	/* Create artificial variables for phase 1.  This ensures that each row has
	a basic variable associated with it.  The original objective row already
	contains the implicit objective variable, which is basic. */
	columns.resize(n_columns+artificial_rows.size());
	columns.back() = columns[n_columns-1];
	columns[n_columns-1].values.clear();
	for(unsigned i=0; i<artificial_rows.size(); ++i)
		columns[n_columns+i-1].basic = artificial_rows[i];
}

void Layout::LinearProgram::remove_artificial_variables()
{
	/* See if any artificial variables are still basic.  This could be because
	the program is degenerate.  To avoid trouble later on, use pivots to make
	some of the original variables basic instead.

	I don't fully understand why this is needed, but it appears to work. */
	for(unsigned i=n_columns-1; i+1<columns.size(); ++i)
		if(columns[i].basic && columns.back().values[columns[i].basic]==0.0f)
		{
			for(unsigned j=0; j+1<n_columns; ++j)
				if(!columns[j].basic && columns[j].values[columns[i].basic]!=0.0f)
				{
					make_basic_column(j, columns[i].basic);
					break;
				}
		}

	/* Get rid of the artificial variables and restore the original objective
	row to form the phase 2 problem. */
	columns.erase(columns.begin()+(n_columns-1), columns.end()-1);
	for(vector<Column>::iterator i=columns.begin(); i!=columns.end(); ++i)
		if(!i->basic)
		{
			i->values.front() = i->values.back();
			i->values.pop_back();
		}
}

unsigned Layout::LinearProgram::find_minimal_ratio(unsigned c)
{
	/* Pick the row with the minimum ratio between the constant column and the
	pivot column.  This ensures that when the pivot column is made basic, values
	in the constant column stay positive.

	The use of n_rows instead of the true size of the column is intentional,
	since the relocated objective row must be ignored in phase 1. */
	float best = numeric_limits<float>::infinity();
	unsigned row = 0;
	for(unsigned i=1; i<n_rows; ++i)
		if(columns[c].values[i]>0)
		{
			float ratio = columns.back().values[i]/columns[c].values[i];
			if(ratio<best)
			{
				best = ratio;
				row = i;
			}
		}

	return row;
}

void Layout::LinearProgram::make_basic_column(unsigned c, unsigned r)
{
	/* Perform row transfer operations to make the pivot column basic,
	containing a 1 on the pivot row. */
	for(unsigned i=0; i<columns.size(); ++i)
		if(i!=c && (columns[i].basic==r || (!columns[i].basic && columns[i].values[r])))
		{
			if(columns[i].basic)
			{
				columns[i].values.resize(columns.back().values.size(), 0.0f);
				columns[i].values[columns[i].basic] = 1.0f;
				columns[i].basic = 0;
			}

			float scale = columns[i].values[r]/columns[c].values[r];

			columns[i].values[r] = scale;
			for(unsigned j=0; j<columns[i].values.size(); ++j)
				if(j!=r)
					columns[i].values[j] -= scale*columns[c].values[j];
		}

	columns[c].basic = r;
	columns[c].values.clear();
}

bool Layout::LinearProgram::pivot()
{
	/* Pick a nonbasic column and make it basic.  Requiring a positive objective
	coefficient ensures that the objective function's value will decrease in the
	process. */
	for(unsigned i=0; i+1<columns.size(); ++i)
		if(!columns[i].basic && columns[i].values.front()>0)
			if(unsigned row = find_minimal_ratio(i))
			{
				make_basic_column(i, row);
				return true;
			}

	return false;
}


Layout::LinearProgram::Row::Row(LinearProgram &lp, unsigned i):
	linprog(lp),
	index(i)
{ }

float &Layout::LinearProgram::Row::operator[](unsigned c)
{
	if(c>=linprog.n_columns)
		throw out_of_range("Row::operator[]");

	Column &column = linprog.columns[c];
	if(column.values.size()<=index)
		column.values.resize(index+1);

	return column.values[index];
}

float &Layout::LinearProgram::Row::back()
{
	return (*this)[linprog.n_columns-1];
}


Layout::LinearProgram::Column::Column():
	basic(0)
{ }

} // namespace GLtk
} // namespace Msp
