#include <msp/gl/matrix.h>
#include <msp/gl/meshbuilder.h>
#include "graphic.h"
#include "list.h"
#include "part.h"
#include "style.h"
#include "text.h"
#include "vslider.h"

using namespace std;

namespace Msp {
namespace GLtk {

List::List():
	data(new BasicListData<string>),
	own_data(true)
{
	init();
}

List::List(ListData &d):
	data(&d),
	own_data(false)
{
	init();
}

void List::init()
{
	sel_index = -1;
	first = 0;
	max_scroll = 0;
	view_size = 5;

	observer = new DataObserver(*this);

	add(slider);
	slider.set_step(1);
	slider.signal_value_changed.connect(sigc::mem_fun(this, &List::slider_value_changed));
}

List::~List()
{
	delete observer;
	if(own_data)
		delete data;
}

void List::autosize()
{
	if(!style)
		return;

	Widget::autosize();

	if(const Part *items_part = style->get_part("items"))
	{
		const Sides &margin = items_part->get_margin();

		unsigned max_w = 0;
		unsigned total_h = 0;
		for(unsigned i=0; i<items.size(); ++i)
		{
			items[i]->autosize();
			const Geometry &igeom = items[i]->get_geometry();
			max_w = max(max_w, igeom.w);
			if(view_size==0 || i<view_size)
				total_h += igeom.h;
		}

		if(!items.empty() && items.size()<view_size)
			total_h = total_h*view_size/items.size();

		geom.w = max(geom.w, max_w+margin.left+margin.right);
		geom.h = max(geom.h, total_h+margin.top+margin.bottom);
	}

	check_view_range();
	rebuild();
}

void List::set_data(ListData &d)
{
	delete observer;
	if(own_data)
		delete data;

	data = &d;
	own_data = false;
	observer = new DataObserver(*this);

	items_changed();
}

void List::items_changed()
{
	check_view_range();
	signal_autosize_changed.emit();
	reposition_items();
}

List::Item *List::create_item(unsigned index)
{
	return new BasicItem(data->get_string(index));
}

void List::set_view_size(unsigned s)
{
	view_size = s;
	signal_autosize_changed.emit();
}

void List::set_view_all()
{
	set_view_size(0);
}

void List::set_selected_index(int i)
{
	if(i>static_cast<int>(data->size()))
		throw out_of_range("List::set_selected_index");

	if(sel_index>=0)
		items[sel_index]->set_active(false);
	if(i<0)
		sel_index = -1;
	else
	{
		sel_index = i;
		items[sel_index]->set_active(true);
		signal_item_selected.emit(sel_index);
	}
}

void List::render_special(const Part &part, GL::Renderer &renderer) const
{
	if(part.get_name()=="items")
	{
		for(unsigned i=first; (i<items.size() && items[i]->is_visible()); ++i)
			items[i]->render(renderer);
	}
	else if(part.get_name()=="slider")
		slider.render(renderer);
}

void List::button_press(int x, int y, unsigned btn)
{
	Container::button_press(x, y, btn);
	if(click_focus && btn==1)
	{
		for(unsigned i=first; (i<items.size() && items[i]->is_visible()); ++i)
			if(click_focus==items[i])
			{
				set_selected_index(i);
				break;
			}
	}
}

void List::on_geometry_change()
{
	reposition_slider();
	reposition_items();

	check_view_range();
}

void List::on_style_change()
{
	if(!style)
		return;

	reposition_slider();
	reposition_items();

	check_view_range();
}

void List::reposition_slider()
{
	if(!style)
		return;

	if(const Part *slider_part = style->get_part("slider"))
	{
		Geometry sgeom = slider_part->get_geometry();
		slider_part->get_alignment().apply(sgeom, geom, slider_part->get_margin());
		slider.set_geometry(sgeom);
	}
}

void List::reposition_items()
{
	if(!style)
		return;

	if(const Part *items_part = style->get_part("items"))
	{
		const Sides &margin = items_part->get_margin();
		unsigned w = geom.w-margin.left-margin.right;
		unsigned y = geom.h-margin.top;
		for(unsigned i=0; i<items.size(); ++i)
		{
			if(i<first || !y)
				items[i]->set_visible(false);
			else
			{
				Geometry igeom = items[i]->get_geometry();
				if(igeom.h+margin.bottom<=y)
				{
					items[i]->set_visible(true);
					y -= igeom.h;
					igeom.x = margin.left;
					igeom.y = y;
					igeom.w = w;
					items[i]->set_geometry(igeom);
				}
				else
				{
					items[i]->set_visible(false);
					y = 0;
				}
			}
		}
	}
}

void List::check_view_range()
{
	if(!style)
		return;

	unsigned h = geom.h;
	if(const Part *items_part = style->get_part("items"))
	{
		const Sides &margin = items_part->get_margin();
		h -= margin.top+margin.bottom;
	}

	max_scroll = items.size();
	for(unsigned i=items.size(); i-->0; )
	{
		unsigned ih = items[i]->get_geometry().h;
		if(ih<=h)
		{
			h -= ih;
			--max_scroll;
		}
	}

	if(first>max_scroll)
		first = max_scroll;

	slider.set_range(0, max_scroll);
	slider.set_value(max_scroll-first);
}

void List::slider_value_changed(double value)
{
	if(max_scroll>0)
	{
		first = max_scroll-static_cast<unsigned>(value);
		reposition_items();
	}
}


List::DataObserver::DataObserver(List &l):
	list(l)
{
	list.data->signal_item_added.connect(sigc::mem_fun(this, &DataObserver::item_added));
	list.data->signal_item_removed.connect(sigc::mem_fun(this, &DataObserver::item_removed));
	list.data->signal_cleared.connect(sigc::mem_fun(this, &DataObserver::cleared));
	list.data->signal_refresh_strings.connect(sigc::mem_fun(this, &DataObserver::refresh_strings));
}

void List::DataObserver::item_added(unsigned i)
{
	if(list.sel_index>=static_cast<int>(i))
		++list.sel_index;

	Item *item = list.create_item(i);
	list.add(*item);
	list.items.insert(list.items.begin()+i, item);
	list.items_changed();
}

void List::DataObserver::item_removed(unsigned i)
{
	if(list.sel_index>static_cast<int>(i))
		--list.sel_index;
	else if(list.sel_index==static_cast<int>(i))
		list.sel_index = -1;

	delete list.items[i];
	list.items.erase(list.items.begin()+i);
	list.items_changed();
}

void List::DataObserver::cleared()
{
	list.sel_index = -1;
	for(vector<Item *>::iterator i=list.items.begin(); i!=list.items.end(); ++i)
		delete *i;
	list.items.clear();
	list.items_changed();
}

void List::DataObserver::refresh_strings()
{
	list.items_changed();
}


void List::Item::autosize()
{
	Widget::autosize();

	if(const Part *part = style->get_part("children"))
	{
		const Sides &margin = part->get_margin();
		for(list<Child *>::const_iterator i=children.begin(); i!=children.end(); ++i)
		{
			const Geometry &cgeom = (*i)->widget->get_geometry();
			geom.w = max(geom.w, cgeom.x+cgeom.w+margin.right);
			geom.h = max(geom.h, cgeom.y+cgeom.h+margin.top);
		}
	}
}

void List::Item::set_active(bool a)
{
	set_state(ACTIVE, (a ? ACTIVE : NORMAL));
}

void List::Item::render_special(const Part &part, GL::Renderer &renderer) const
{
	if(part.get_name()=="children")
	{
		for(list<Child *>::const_iterator i=children.begin(); i!=children.end(); ++i)
			(*i)->widget->render(renderer);
	}
}


List::BasicItem::BasicItem(const string &text):
	label(text)
{
	add(label);
}

void List::BasicItem::on_style_change()
{
	if(!style)
		return;

	label.autosize();
	if(const Part *part = style->get_part("children"))
	{
		const Sides &margin = part->get_margin();
		label.set_position(margin.left, margin.bottom);
	}
}


List::Loader::Loader(List &l):
	DataFile::DerivedObjectLoader<List, Widget::Loader>(l)
{
	add("item", &Loader::item);
}

void List::Loader::item(const string &v)
{
	dynamic_cast<BasicListData<string> &>(*obj.data).append(v);
}

} // namespace GLtk
} // namespace Msp
