#include <msp/gl/matrix.h>
#include <msp/gl/meshbuilder.h>
#include <msp/gl/texture2d.h>
#include "partcache.h"
#include "style.h"
#include "text.h"

using namespace std;

namespace Msp {
namespace GLtk {

struct Text::RenderData
{
	GL::PrimitiveBuilder *bld;
};

struct Text::CoordsToGeomData
{
	unsigned row;
	unsigned col;
	Geometry result;
};


Text::Text():
	style(0)
{ }

Text::Text(const Style &s, const string &t):
	style(&s)
{
	set(t);
}

void Text::set_style(const Style *s)
{
	style = s;

	if(style)
	{
		const GL::Font &font = style->get_font();
		float font_size = style->get_font_size();
		for(vector<Line>::iterator i=lines.begin(); i!=lines.end(); ++i)
			i->width = static_cast<unsigned>(font.get_string_width(text.substr(i->start, i->length))*font_size);
	}
}

unsigned Text::get_width() const
{
	unsigned width = 0;
	for(vector<Line>::const_iterator i=lines.begin(); i!=lines.end(); ++i)
		width = max(width, i->width);
	return width;
}

unsigned Text::get_height() const
{
	if(!style)
		return lines.size();

	const GL::Font &font = style->get_font();
	float font_size = style->get_font_size();
	unsigned line_height = static_cast<unsigned>((font.get_ascent()-font.get_descent())*font_size);
	unsigned line_spacing = line_height*6/5;
	return line_height+(lines.size()-1)*line_spacing;
}

void Text::set(const string &t)
{
	text = t;
	find_lines();
}

void Text::erase(unsigned pos, unsigned len)
{
	text.erase(pos, len);

	vector<Line>::iterator i;
	for(i=lines.begin(); (i!=lines.end() && i->start+i->length<pos); ++i) ;

	if(pos+len>i->start+i->length)
		find_lines();
	else
	{
		i->length -= len;

		for(++i; i!=lines.end(); ++i)
			i->start -= len;
	}
}

void Text::insert(unsigned pos, const string &s)
{
	text.insert(pos, s);

	if(s.find('\n')!=string::npos)
		find_lines();
	else
	{
		vector<Line>::iterator i;
		for(i=lines.begin(); (i!=lines.end() && i->start+i->length<pos); ++i) ;

		i->length += s.size();

		for(++i; i!=lines.end(); ++i)
			i->start += s.size();
	}
}

unsigned Text::get_line_length(unsigned i) const
{
	if(i>=lines.size())
		throw out_of_range("Text::get_line_length");
	return lines[i].length;
}

void Text::offset_to_coords(unsigned offs, unsigned &row, unsigned &col) const
{
	if(lines.empty())
	{
		row = 0;
		col = 0;
		return;
	}

	for(unsigned i=0; i<lines.size(); ++i)
		if(offs>=lines[i].start && offs<=lines[i].start+lines[i].length)
		{
			row = i;
			col = offs-lines[i].start;
			return;
		}
}

unsigned Text::coords_to_offset(unsigned row, unsigned col) const
{
	if(row>=lines.size())
		return text.size();

	return lines[row].start+min(col, lines[row].length);
}

Geometry Text::coords_to_geometry(const Part &part, const Geometry &parent, unsigned first_row, unsigned row, unsigned col) const
{
	if(row>=lines.size())
		row = lines.size()-1;
	const Line &line = lines[row];
	if(col>line.length)
		col = line.length;

	CoordsToGeomData data;
	data.row = row;
	data.col = col;

	process_lines<CoordsToGeomData, &Text::coords_to_geom_line>(part, parent, first_row, data);

	return data.result;
}

void Text::build(const Part &part, const Geometry &parent, PartCache &cache) const
{
	build(part, parent, 0, cache);
}

void Text::build(const Part &part, const Geometry &parent, unsigned first_row, PartCache &cache) const
{
	if(!style || lines.empty())
		return;

	const GL::Font &font = style->get_font();
	GL::MeshBuilder bld(cache.create_mesh(part, font.get_texture()));
	bld.color(style->get_font_color());

	RenderData data;
	data.bld = &bld;

	process_lines<RenderData, &Text::build_line>(part, parent, first_row, data);
}

Text &Text::operator=(const string &t)
{
	set(t);
	return *this;
}

void Text::find_lines()
{
	lines.clear();
	float font_size = (style ? style->get_font_size() : 1);
	string::size_type start = 0;
	while(1)
	{
		string::size_type newline = text.find('\n', start);

		Line line;
		line.start = start;
		line.length = (newline==string::npos ? text.size() : newline)-start;
		line.width = line.length;
		if(style)
		{
			string str = text.substr(line.start, line.length);
			line.width = static_cast<unsigned>(style->get_font().get_string_width(str)*font_size);
		}
		lines.push_back(line);

		if(newline==string::npos)
			break;
		start = newline+1;
	}
}

template<typename T, void (Text::*func)(unsigned, const Geometry &, T &) const>
void Text::process_lines(const Part &part, const Geometry &parent, unsigned first_row, T &data) const
{
	if(!style)
		return;

	const GL::Font &font = style->get_font();
	float font_size = style->get_font_size();
	unsigned line_height = static_cast<unsigned>((font.get_ascent()-font.get_descent())*font_size);
	unsigned line_spacing = static_cast<unsigned>(font_size*6/5);
	unsigned height = line_height+(lines.size()-1)*line_spacing;
	int y_offset = static_cast<int>(-font.get_descent()*font_size);

	const Sides &margin = part.get_margin();
	unsigned n_lines = min<unsigned>(lines.size(), max((parent.h-margin.top-margin.bottom)/line_spacing, 1U));
	first_row = min<unsigned>(first_row, lines.size()-n_lines);

	for(unsigned i=0; i<n_lines; ++i)
	{
		const Line &line = lines[first_row+i];

		Geometry rgeom;
		rgeom.w = line.width;
		rgeom.h = height;
		rgeom.y = (n_lines-1-i)*line_spacing+y_offset;
		part.get_alignment().apply(rgeom, parent, part.get_margin());

		(this->*func)(first_row+i, rgeom, data);
	}
}

void Text::build_line(unsigned i, const Geometry &rgeom, RenderData &data) const
{
	const Line &line = lines[i];

	GL::MatrixStack::Push _pushm(data.bld->matrix());
	data.bld->matrix() *= GL::Matrix::translation(rgeom.x, rgeom.y, 0);
	data.bld->matrix() *= GL::Matrix::scaling(style->get_font_size());

	style->get_font().build_string(text.substr(line.start, line.length), *data.bld);
}

void Text::coords_to_geom_line(unsigned i, const Geometry &rgeom, CoordsToGeomData &data) const
{
	if(i==data.row)
	{
		float w = style->get_font().get_string_width(text.substr(lines[i].start, data.col));
		data.result = rgeom;
		data.result.x += static_cast<unsigned>(w*style->get_font_size());
	}
}

} // namespace GLtk
} // namespace Msp
